@@ -70,6 +70,12 @@ int compile(struct compilation_unit *cu)
 	if (err)
 		goto out;
 
+	if (opt_ssa_enable) {
+		err = compute_dfns(cu);
+		if (err)
+			goto out;
+	}
+
 	if (opt_trace_cfg)
 		trace_cfg(cu);
 
@@ -85,6 +91,16 @@ int compile(struct compilation_unit *cu)
 	if (opt_trace_lir)
 		trace_lir(cu);
 
+	if (opt_ssa_enable) {
+		err = compute_dom(cu);
+		if (err)
+			goto out;
+
+		err = compute_dom_frontier(cu);
+		if (err)
+			goto out;
+	}
+
 	err = analyze_liveness(cu);
 	if (err)
 		goto out;
@@ -119,14 +135,13 @@ int compile(struct compilation_unit *cu)
 	resolve_fixup_offsets(cu);
 
 	cu->native_ptr = buffer_ptr(cu->objcode);
-	cu->is_compiled = true;
 
 	perf_append_cu(cu);
   out:
 	if (opt_trace_compile)
 		trace_flush();
 
-	if (err)
+	if (err && !exception_occurred())
 		compile_error(cu, err);
 
 	return err;
