@@ -38,6 +38,7 @@
 #include "jit/cu-mapping.h"
 
 #include "lib/guard-page.h"
+#include "lib/string.h"
 
 #include "vm/stdlib.h"
 #include "vm/thread.h"
@@ -46,7 +47,6 @@
 #include "vm/trace.h"
 #include "vm/die.h"
 #include "vm/gc.h"
-#include "lib/string.h"
 
 #include <inttypes.h>
 #include <pthread.h>
@@ -54,8 +54,6 @@
 #include <assert.h>
 #include <stdio.h>
 
-#include "../boehmgc/include/gc.h"
-
 void *gc_safepoint_page;
 
 static pthread_mutex_t	gc_reclaim_mutex	= PTHREAD_MUTEX_INITIALIZER;
@@ -73,8 +71,12 @@ static __thread sig_atomic_t in_safepoint;
 static pthread_t gc_thread_id;
 
 unsigned long max_heap_size	= 128 * 1024 * 1024;	/* 128 MB */
-bool verbose_gc;
-int dont_gc;
+
+bool				newgc_enabled;
+bool				verbose_gc;
+int				dont_gc;
+
+struct gc_operations		gc_ops;
 
 static void hide_safepoint_guard_page(void)
 {
@@ -205,12 +207,12 @@ void suspend_handler(int sig, siginfo_t *si, void *ctx)
 		 * Fresh threads might return NULL from vm_thread_self().
 		 */
 		if (self)
-			self->thread_state = THREAD_STATE_CONSISTENT;
+			vm_thread_set_state(self, THREAD_STATE_CONSISTENT);
 
 		save_signal_registers(&thread_register_state, uc->uc_mcontext.gregs);
 		gc_safepoint(&thread_register_state);
 	} else {
-		self->thread_state = THREAD_STATE_INCONSISTENT;
+		vm_thread_set_state(self, THREAD_STATE_CONSISTENT);
 
 		enter_safepoint();
 
@@ -286,7 +288,7 @@ static void gc_suspend_rest(void)
 	vm_thread_for_each(thread) {
 		assert(thread->posix_id != pthread_self());
 
-		if (thread->thread_state == THREAD_STATE_INCONSISTENT) {
+		if (vm_thread_get_state(thread) == THREAD_STATE_INCONSISTENT) {
 			resume_thread(thread->posix_id);
 			nr_restarted++;
 		}
@@ -407,42 +409,55 @@ wait_for_reclaim:
 		die("pthread_mutex_unlock");
 }
 
-static void *gc_out_of_memory(size_t nr)
+static void *do_gc_alloc(size_t size)
 {
-	if (verbose_gc)
-		fprintf(stderr, "[GC]\n");
+	void *p;
 
-	GC_gcollect();
+	gc_start();
 
-	return NULL;	/* is this ok? */
+	p	= malloc(size);
+
+	return p;
 }
 
-static void gc_ignore_warnings(char *msg, GC_word arg)
+static void *do_vm_alloc(size_t size)
 {
+	return malloc(size);
 }
 
-static void setup_boehm_gc(void)
+void *vm_zalloc(size_t size)
 {
-	GC_set_warn_proc(gc_ignore_warnings);
+	void *ret;
 
-	GC_oom_fn	= gc_out_of_memory;
+	ret = vm_alloc(size);
+	if (!ret)
+		return NULL;
 
-	GC_quiet	= 1;
+	memset(ret, 0, size);
 
-	GC_dont_gc	= dont_gc;
-
-	GC_INIT();
-
-        GC_set_max_heap_size(max_heap_size);
+	return ret;
 }
 
-void gc_init(void)
+static void do_vm_free(void *p)
 {
-	setup_boehm_gc();
+	free(p);
+}
 
-	gc_safepoint_page = alloc_guard_page(false);
-	if (!gc_safepoint_page)
-		die("Couldn't allocate GC safepoint guard page");
+static int do_gc_register_finalizer(struct vm_object *object,
+				    finalizer_fn finalizer)
+{
+	NOT_IMPLEMENTED;
+	return 0;
+}
+
+static void gc_setup(void)
+{
+	gc_ops		= (struct gc_operations) {
+		.gc_alloc		= do_gc_alloc,
+		.vm_alloc		= do_vm_alloc,
+		.vm_free		= do_vm_free,
+		.gc_register_finalizer	= do_gc_register_finalizer
+	};
 
 	if (pthread_spin_init(&gc_spinlock, PTHREAD_PROCESS_SHARED) != 0)
 		die("pthread_spin_init");
@@ -451,33 +466,14 @@ void gc_init(void)
 		die("Couldn't create GC thread");
 }
 
-void *gc_alloc(size_t size)
+void gc_init(void)
 {
-	return GC_MALLOC(size);
-}
+	gc_safepoint_page = alloc_guard_page(false);
+	if (!gc_safepoint_page)
+		die("Couldn't allocate GC safepoint guard page");
 
-void *vm_alloc(size_t size)
-{
-	return GC_malloc_uncollectable(size);
-}
-
-void *vm_zalloc(size_t size)
-{
-	void *result;
-
-	result = vm_alloc(size);
-	memset(result, 0, size);
-	return result;
-}
-
-void vm_free(void *ptr)
-{
-	GC_free(ptr);
-}
-
-void gc_register_finalizer(struct vm_object *object, finalizer_fn finalizer,
-			   void *param)
-{
-	GC_register_finalizer(object, (GC_finalization_proc) finalizer,
-			      param, NULL, NULL);
+	if (newgc_enabled)
+		gc_setup();
+	else
+		gc_setup_boehm();
 }
