@@ -25,7 +25,6 @@
  * Please refer to the file LICENSE for details.
  */
 
-#include <valgrind/valgrind.h>
 #include <sys/utsname.h>
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -40,6 +39,8 @@
 #include <stdio.h>
 #include <time.h>
 
+#include "valgrind/valgrind.h"
+
 #include "cafebabe/access.h"
 #include "cafebabe/attribute_info.h"
 #include "cafebabe/class.h"
@@ -57,10 +58,12 @@
 
 #include "jit/compiler.h"
 #include "jit/cu-mapping.h"
+#include "jit/gdb.h"
 #include "jit/exception.h"
 #include "jit/perf-map.h"
 #include "jit/text.h"
 
+#include "lib/parse.h"
 #include "lib/list.h"
 
 #include "vm/fault-inject.h"
@@ -73,6 +76,7 @@
 #include "vm/itable.h"
 #include "vm/method.h"
 #include "vm/object.h"
+#include "vm/reference.h"
 #include "vm/signal.h"
 #include "vm/static.h"
 #include "vm/string.h"
@@ -87,6 +91,7 @@
 
 #include "arch/init.h"
 
+static bool dump_maps;
 static bool perf_enabled;
 static char *exe_name;
 
@@ -94,6 +99,11 @@ static char *exe_name;
 static unsigned int nr_java_args;
 static char **java_args;
 
+/*
+ * Enable SSA optimizations.
+ */
+bool opt_ssa_enable;
+
 /*
  * Enable JIT workarounds for valgrind.
  */
@@ -199,26 +209,40 @@ static void system_property_append_path(const char *key, const char *path)
 		error("out of memory");
 }
 
+static char *system_property_get(const char *key)
+{
+	struct system_properties_entry *ent;
+
+	ent = find_system_property(key);
+	if (!ent)
+		return NULL;
+
+	return ent->value;
+}
+
 struct system_property {
 	const char *key;
 	const char *value;
 };
 
+#define JAVA_VERSION		"1.5.0"
+
 static struct system_property system_properties[] = {
 	{ "java.vm.name",			"jato"				},
 	{ "java.vm.vendor",			"Pekka Enberg"			},
 	{ "java.vm.vendor.url",			"http://jatovm.sourceforge.net/"},
 	{ "java.io.tmpdir",			"/tmp"				},
+	{ "java.home",				"/usr"				},
 	{ "file.separator",			"/"				},
 	{ "path.separator",			":"				},
 	{ "line.separator",			"\n"				},
-	{ "java.version",			"1.6"				},
+	{ "java.version",			JAVA_VERSION			},
 	{ "java.vendor",			"GNU Classpath"			},
 	{ "java.vendor.url",			"http://www.classpath.org"	},
 	{ "java.vm.specification.version",	"1.0"				},
 	{ "java.vm.specification.vendor",	"Sun Microsystems, Inc."	},
 	{ "java.vm.specification.name",		"Java Virtual Machine Specification"},
-	{ "java.specification.version",		"1.6"				},
+	{ "java.specification.version",		"1.5"				},
 	{ "java.specification.vendor",		"Sun Microsystems, Inc."	},
 	{ "java.specification.name",		"Java Platform API Specification"},
 	{ "java.class.version",			"48.0"				},
@@ -261,19 +285,6 @@ static void init_system_properties(void)
 	add_system_property_const("gnu.cpu.endian", "little");
 }
 
-
-/**
- * This should be called after parsing of call arguments.
- */
-static void init_configurable_system_properties(void)
-{
-	/* XXX: currently user defined -Djava.class.path=value is overriden. */
-	char *cp = get_classpath();
-	add_system_property_const("java.class.path", cp);
-	add_system_property_const("java.boot.class.path", cp);
-	free(cp);
-}
-
 static void native_vmsystemproperties_preinit(struct vm_object *p)
 {
 	struct system_properties_entry *this, *t;
@@ -339,16 +350,16 @@ native_vmsystem_arraycopy(struct vm_object *src, int src_start,
 	}
 
 	if (len < 0 ||
-	    src_start < 0 || src_start + len > src->array_length ||
-	    dest_start < 0 || dest_start + len > dest->array_length) {
+	    src_start < 0 || src_start + len > vm_array_length(src) ||
+	    dest_start < 0 || dest_start + len > vm_array_length(dest)) {
 		signal_new_exception(
 			vm_java_lang_ArrayIndexOutOfBoundsException, NULL);
 		return;
 	}
 
-	elem_size = get_vmtype_size(elem_type);
-	memmove(dest->fields + dest_start * elem_size,
-		src->fields + src_start * elem_size,
+	elem_size = vmtype_get_size(elem_type);
+	memmove(vm_array_elems(dest) + dest_start * elem_size,
+		vm_array_elems(src) + src_start * elem_size,
 		len * elem_size);
 
 	return;
@@ -443,6 +454,11 @@ static void native_vmthread_start(struct vm_object *vmthread, jlong stacksize)
 	vm_thread_start(vmthread);
 }
 
+static void native_vmthread_yield(void)
+{
+	vm_thread_yield();
+}
+
 static jint native_vmthread_interrupted(void)
 {
 	return vm_thread_interrupted(vm_thread_self());
@@ -452,10 +468,15 @@ static jint native_vmthread_isinterrupted(struct vm_object *vmthread)
 {
 	struct vm_thread *thread;
 
-	thread = (struct vm_thread *)field_get_object(vmthread, vm_java_lang_VMThread_vmdata);
+	thread = vm_thread_from_vmthread(vmthread);
 	return vm_thread_is_interrupted(thread);
 }
 
+static void native_vmthread_setpriority(struct vm_object *vmthread, jint ptiority)
+{
+	/* TODO: implement */
+}
+
 static void native_vmthread_interrupt(struct vm_object *vmthread)
 {
 	struct vm_thread *thread;
@@ -466,7 +487,7 @@ static void native_vmthread_interrupt(struct vm_object *vmthread)
 
 static void native_vmobject_notify(struct vm_object *obj)
 {
-	vm_monitor_notify(&obj->monitor);
+	vm_object_notify(obj);
 
 	if (exception_occurred())
 		return;
@@ -474,30 +495,18 @@ static void native_vmobject_notify(struct vm_object *obj)
 
 static void native_vmobject_notify_all(struct vm_object *obj)
 {
-	vm_monitor_notify_all(&obj->monitor);
+	vm_object_notify_all(obj);
 
 	if (exception_occurred())
 		return;
 }
 
-static jlong native_vmsystem_nano_time(void)
-{
-	struct timespec time;
-	if (clock_gettime(CLOCK_MONOTONIC, &time)) {
-		NOT_IMPLEMENTED;
-		return 0;
-	}
-
-	return (unsigned long long)time.tv_sec * 1000000000ull +
-		(unsigned long long)time.tv_nsec;
-}
-
 static void native_vmobject_wait(struct vm_object *object, jlong ms, jint ns)
 {
 	if (ms == 0 && ns == 0)
-		vm_monitor_wait(&object->monitor);
+		vm_object_wait(object);
 	else
-		vm_monitor_timed_wait(&object->monitor, ms, ns);
+		vm_object_timed_wait(object, ms, ns);
 }
 
 static struct vm_object *native_vmstring_intern(struct vm_object *str)
@@ -519,6 +528,7 @@ static struct vm_native natives[] = {
 	DEFINE_NATIVE("jato/internal/VM", "exit", native_vmruntime_exit),
 	DEFINE_NATIVE("jato/internal/VM", "println", native_vmruntime_println),
 	DEFINE_NATIVE("jato/internal/VM", "throwNullPointerException", native_vm_throw_null_pointer_exception),
+	DEFINE_NATIVE("java/lang/reflect/VMArray", "createObjectArray", native_vmarray_createobjectarray),
 	DEFINE_NATIVE("java/io/VMFile", "isDirectory", native_vmfile_is_directory),
 	DEFINE_NATIVE("java/lang/VMClass", "forName", native_vmclass_forname),
 	DEFINE_NATIVE("java/lang/VMClass", "getClassLoader", native_vmclass_getclassloader),
@@ -526,6 +536,7 @@ static struct vm_native natives[] = {
 	DEFINE_NATIVE("java/lang/VMClass", "getDeclaredConstructors", native_vmclass_get_declared_constructors),
 	DEFINE_NATIVE("java/lang/VMClass", "getDeclaredFields", native_vmclass_get_declared_fields),
 	DEFINE_NATIVE("java/lang/VMClass", "getDeclaredMethods", native_vmclass_get_declared_methods),
+	DEFINE_NATIVE("java/lang/VMClass", "getDeclaringClass", native_vmclass_get_declaring_class),
 	DEFINE_NATIVE("java/lang/VMClass", "getInterfaces", native_vmclass_get_interfaces),
 	DEFINE_NATIVE("java/lang/VMClass", "getModifiers", native_vmclass_getmodifiers),
 	DEFINE_NATIVE("java/lang/VMClass", "getName", native_vmclass_getname),
@@ -540,43 +551,71 @@ static struct vm_native natives[] = {
 	DEFINE_NATIVE("java/lang/VMClassLoader", "findLoadedClass", native_vmclassloader_findloadedclass),
 	DEFINE_NATIVE("java/lang/VMClassLoader", "getPrimitiveClass", native_vmclassloader_getprimitiveclass),
 	DEFINE_NATIVE("java/lang/VMClassLoader", "loadClass", native_vmclassloader_loadclass),
+	DEFINE_NATIVE("java/lang/VMClassLoader", "resolveClass", native_vmclassloader_resolveclass),
 	DEFINE_NATIVE("java/lang/VMObject", "clone", native_vmobject_clone),
 	DEFINE_NATIVE("java/lang/VMObject", "getClass", native_vmobject_getclass),
 	DEFINE_NATIVE("java/lang/VMObject", "notify", native_vmobject_notify),
 	DEFINE_NATIVE("java/lang/VMObject", "notifyAll", native_vmobject_notify_all),
 	DEFINE_NATIVE("java/lang/VMObject", "wait", native_vmobject_wait),
+	DEFINE_NATIVE("java/lang/VMRuntime", "freeMemory", native_vmruntime_free_memory),
+	DEFINE_NATIVE("java/lang/VMRuntime", "totalMemory", native_vmruntime_total_memory),
+	DEFINE_NATIVE("java/lang/VMRuntime", "maxMemory", native_vmruntime_max_memory),
+	DEFINE_NATIVE("java/lang/VMRuntime", "availableProcessors", native_vmruntime_available_processors),
 	DEFINE_NATIVE("java/lang/VMRuntime", "exit", native_vmruntime_exit),
+	DEFINE_NATIVE("java/lang/VMRuntime", "gc", native_vmruntime_gc),
 	DEFINE_NATIVE("java/lang/VMRuntime", "mapLibraryName", native_vmruntime_maplibraryname),
 	DEFINE_NATIVE("java/lang/VMRuntime", "nativeLoad", native_vmruntime_native_load),
 	DEFINE_NATIVE("java/lang/VMRuntime", "runFinalizationForExit", native_vmruntime_run_finalization_for_exit),
 	DEFINE_NATIVE("java/lang/VMString", "intern", native_vmstring_intern),
 	DEFINE_NATIVE("java/lang/VMSystem", "arraycopy", native_vmsystem_arraycopy),
 	DEFINE_NATIVE("java/lang/VMSystem", "identityHashCode", native_vmsystem_identityhashcode),
-	DEFINE_NATIVE("java/lang/VMSystem", "nanoTime", native_vmsystem_nano_time),
 	DEFINE_NATIVE("java/lang/VMThread", "currentThread", native_vmthread_current_thread),
 	DEFINE_NATIVE("java/lang/VMThread", "interrupt", native_vmthread_interrupt),
 	DEFINE_NATIVE("java/lang/VMThread", "interrupted", native_vmthread_interrupted),
 	DEFINE_NATIVE("java/lang/VMThread", "isInterrupted", native_vmthread_isinterrupted),
+	DEFINE_NATIVE("java/lang/VMThread", "nativeSetPriority", native_vmthread_setpriority),
 	DEFINE_NATIVE("java/lang/VMThread", "start", native_vmthread_start),
+	DEFINE_NATIVE("java/lang/VMThread", "yield", native_vmthread_yield),
 	DEFINE_NATIVE("java/lang/VMThrowable", "fillInStackTrace", native_vmthrowable_fill_in_stack_trace),
 	DEFINE_NATIVE("java/lang/VMThrowable", "getStackTrace", native_vmthrowable_get_stack_trace),
 	DEFINE_NATIVE("java/lang/reflect/Constructor", "constructNative", native_constructor_construct_native),
 	DEFINE_NATIVE("java/lang/reflect/Constructor", "getModifiersInternal", native_constructor_get_modifiers_internal),
 	DEFINE_NATIVE("java/lang/reflect/Constructor", "getParameterTypes", native_constructor_get_parameter_types),
+	DEFINE_NATIVE("java/lang/reflect/Constructor", "getExceptionTypes", native_vmconstructor_get_exception_types),
+	DEFINE_NATIVE("java/lang/reflect/VMConstructor", "construct", native_vmconstructor_construct),
+	DEFINE_NATIVE("java/lang/reflect/VMConstructor", "getParameterTypes", native_constructor_get_parameter_types),
+	DEFINE_NATIVE("java/lang/reflect/VMConstructor", "getModifiersInternal", native_constructor_get_modifiers_internal),
+	DEFINE_NATIVE("java/lang/reflect/VMConstructor", "getExceptionTypes", native_vmconstructor_get_exception_types),
 	DEFINE_NATIVE("java/lang/reflect/Field", "get", native_field_get),
+	DEFINE_NATIVE("java/lang/reflect/Field", "getInt", native_field_get_int),
 	DEFINE_NATIVE("java/lang/reflect/Field", "getLong", native_field_get_long),
 	DEFINE_NATIVE("java/lang/reflect/Field", "getModifiersInternal", native_field_get_modifiers_internal),
 	DEFINE_NATIVE("java/lang/reflect/Field", "getType", native_field_gettype),
 	DEFINE_NATIVE("java/lang/reflect/Field", "set", native_field_set),
+	DEFINE_NATIVE("java/lang/reflect/VMField", "get", native_field_get),
+	DEFINE_NATIVE("java/lang/reflect/VMField", "getLong", native_field_get_long),
+	DEFINE_NATIVE("java/lang/reflect/VMField", "getInt", native_field_get_int),
+	DEFINE_NATIVE("java/lang/reflect/VMField", "getModifiersInternal", native_field_get_modifiers_internal),
+	DEFINE_NATIVE("java/lang/reflect/VMField", "getType", native_field_gettype),
+	DEFINE_NATIVE("java/lang/reflect/VMField", "set", native_field_set),
 	DEFINE_NATIVE("java/lang/reflect/Method", "getModifiersInternal", native_method_get_modifiers_internal),
 	DEFINE_NATIVE("java/lang/reflect/Method", "getParameterTypes", native_method_get_parameter_types),
 	DEFINE_NATIVE("java/lang/reflect/Method", "getReturnType", native_method_getreturntype),
 	DEFINE_NATIVE("java/lang/reflect/Method", "invokeNative", native_method_invokenative),
+	DEFINE_NATIVE("java/lang/reflect/VMMethod", "getExceptionTypes", native_method_get_exception_types),
+	DEFINE_NATIVE("java/lang/reflect/VMMethod", "getModifiersInternal", native_method_get_modifiers_internal),
+	DEFINE_NATIVE("java/lang/reflect/VMMethod", "getParameterTypes", native_method_get_parameter_types),
+	DEFINE_NATIVE("java/lang/reflect/VMMethod", "invoke", native_vmmethod_invoke),
+	DEFINE_NATIVE("java/lang/reflect/VMMethod", "getReturnType", native_method_getreturntype),
+	DEFINE_NATIVE("java/lang/reflect/VMMethod", "getExceptionTypes", native_method_get_exception_types),
 	DEFINE_NATIVE("java/util/concurrent/atomic/AtomicLong", "VMSupportsCS8", native_atomiclong_vm_supports_cs8),
 	DEFINE_NATIVE("sun/misc/Unsafe", "compareAndSwapInt", native_unsafe_compare_and_swap_int),
 	DEFINE_NATIVE("sun/misc/Unsafe", "compareAndSwapLong", native_unsafe_compare_and_swap_long),
 	DEFINE_NATIVE("sun/misc/Unsafe", "compareAndSwapObject", native_unsafe_compare_and_swap_object),
 	DEFINE_NATIVE("sun/misc/Unsafe", "objectFieldOffset", native_unsafe_object_field_offset),
+	DEFINE_NATIVE("sun/misc/Unsafe", "putObject", native_unsafe_put_object),
+	DEFINE_NATIVE("sun/misc/Unsafe", "park", native_unsafe_park),
+	DEFINE_NATIVE("sun/misc/Unsafe", "unpark", native_unsafe_unpark),
 };
 
 static void jit_init_natives(void)
@@ -595,6 +634,7 @@ static void usage(FILE *f, int retval)
 
 static void handle_version(void)
 {
+	fprintf(stdout, "java version \"%s\"\n", JAVA_VERSION);
 	fprintf(stdout, "Jato VM version %s\n", JATO_VERSION);
 	exit(EXIT_SUCCESS);
 }
@@ -606,7 +646,22 @@ static void handle_help(void)
 
 static void handle_classpath(const char *arg)
 {
-	classloader_add_to_classpath(arg);
+	system_property_append_path("java.class.path", arg);
+}
+
+static void handle_bootclasspath(const char *arg)
+{
+	static const char delim[] = ":;";
+
+	add_system_property_const("java.boot.class.path", arg);
+
+	char *cp = strdup(arg);
+
+	char *path = strtok(cp, delim);
+	while (path) {
+		classloader_add_to_classpath(path);
+		path = strtok(NULL, delim);
+	}
 }
 
 enum operation {
@@ -646,12 +701,22 @@ static void handle_jar(const char *arg)
 	/* XXX: Cheap solution. This can give funny results depending on where
 	 * you put the -jar relative to the -classpath(s). Besides, we should
 	 * save some memory and only open the zip file once. */
-	classloader_add_to_classpath(arg);
+	system_property_append_path("java.class.path", arg);
 }
 
-static void handle_gc(void)
+static void handle_nogc(void)
 {
-	gc_enabled = true;
+	dont_gc		= 1;
+}
+
+static void handle_newgc(void)
+{
+	newgc_enabled	= true;
+}
+
+static void handle_maps(void)
+{
+	dump_maps = true;
 }
 
 static void handle_perf(void)
@@ -659,16 +724,18 @@ static void handle_perf(void)
 	perf_enabled = true;
 }
 
-/* @arg must be in the format package/name/Class.method(Lsignature;)V */
-static void handle_trace_method(const char *arg)
+static void handle_ssa(void)
 {
-	opt_trace_method = true;
+	opt_ssa_enable = true;
+}
 
-	int err = regcomp(&method_trace_regex, arg, REG_EXTENDED | REG_NOSUB);
+static void regex_compile(regex_t *regex, const char *arg)
+{
+	int err = regcomp(regex, arg, REG_EXTENDED | REG_NOSUB);
 	if (err) {
-		unsigned int size = regerror(err, &method_trace_regex, NULL, 0);
+		unsigned int size = regerror(err, regex, NULL, 0);
 		char *errbuf = malloc(size);
-		regerror(err, &method_trace_regex, errbuf, size);
+		regerror(err, regex, errbuf, size);
 
 		fprintf(stderr, "error: regcomp: %s\n", errbuf);
 		free(errbuf);
@@ -677,6 +744,19 @@ static void handle_trace_method(const char *arg)
 	}
 }
 
+/* @arg must be in the format package/name/Class.method(Lsignature;)V */
+static void handle_trace_method(const char *arg)
+{
+	opt_trace_method = true;
+	regex_compile(&method_trace_regex, arg);
+}
+
+static void handle_trace_gate(const char *arg)
+{
+	opt_trace_gate = true;
+	regex_compile(&method_trace_gate_regex, arg);
+}
+
 static void handle_trace_asm(void)
 {
 	opt_trace_machine_code = true;
@@ -777,6 +857,16 @@ static void handle_verbose_gc(void)
 	verbose_gc = true;
 }
 
+static void handle_max_heap_size(const char *arg)
+{
+	max_heap_size = parse_long(arg);
+
+	if (!max_heap_size) {
+		fprintf(stderr, "%s: unparseable heap size '%s'\n", exe_name, arg);
+		usage(stderr, EXIT_FAILURE);
+	}
+}
+
 struct option {
 	const char *name;
 
@@ -799,21 +889,27 @@ struct option {
 	{ .name = _name, .arg = true, .arg_is_adjacent = true, .handler.func_arg = _handler }
 
 const struct option options[] = {
-	DEFINE_OPTION("version",	handle_version),
-	DEFINE_OPTION("h",		handle_help),
-	DEFINE_OPTION("help",		handle_help),
+	DEFINE_OPTION("version",		handle_version),
+	DEFINE_OPTION("h",			handle_help),
+	DEFINE_OPTION("help",			handle_help),
 
-	DEFINE_OPTION_ADJACENT_ARG("D",	handle_define),
+	DEFINE_OPTION_ADJACENT_ARG("D",		handle_define),
+	DEFINE_OPTION_ADJACENT_ARG("Xmx",	handle_max_heap_size),
 
-	DEFINE_OPTION_ARG("classpath",	handle_classpath),
-	DEFINE_OPTION_ARG("cp",		handle_classpath),
-	DEFINE_OPTION_ARG("jar",	handle_jar),
-	DEFINE_OPTION("verbose:gc",	handle_verbose_gc),
+	DEFINE_OPTION_ARG("classpath",		handle_classpath),
+	DEFINE_OPTION_ARG("bootclasspath",	handle_bootclasspath),
+	DEFINE_OPTION_ARG("cp",			handle_classpath),
+	DEFINE_OPTION_ARG("jar",		handle_jar),
+	DEFINE_OPTION("verbose:gc",		handle_verbose_gc),
 
-	DEFINE_OPTION("Xgc",			handle_gc),
+	DEFINE_OPTION("Xmaps",			handle_maps),
+	DEFINE_OPTION("Xnewgc",			handle_newgc),
+	DEFINE_OPTION("Xnogc",			handle_nogc),
 	DEFINE_OPTION("Xperf",			handle_perf),
+	DEFINE_OPTION("Xssa",			handle_ssa),
 
 	DEFINE_OPTION_ARG("Xtrace:method",	handle_trace_method),
+	DEFINE_OPTION_ARG("Xtrace:gate",	handle_trace_gate),
 
 	DEFINE_OPTION("Xtrace:asm",		handle_trace_asm),
 	DEFINE_OPTION("Xtrace:bytecode",	handle_trace_bytecode),
@@ -853,8 +949,10 @@ static void parse_options(int argc, char *argv[])
 			break;
 
 		const struct option *opt = get_option(argv[optind] + 1);
-		if (!opt)
+		if (!opt) {
+			fprintf(stderr, "%s: unrecognized option '%s'\n", exe_name, argv[optind]);
 			usage(stderr, EXIT_FAILURE);
+		}
 
 		if (!opt->arg) {
 			opt->handler.func();
@@ -872,6 +970,11 @@ static void parse_options(int argc, char *argv[])
 			usage(stderr, EXIT_FAILURE);
 
 		opt->handler.func_arg(argv[++optind]);
+
+		if (strcmp(argv[optind - 1], "-jar") == 0) {
+			optind++;
+			break;
+		}
 	}
 
 	if (operation == OPERATION_MAIN_CLASS) {
@@ -904,6 +1007,10 @@ do_main_class(void)
 	if (!loader)
 		return -1;
 
+	if (exception_occurred()) {
+		return -1;
+	}
+
 	struct vm_class *vmc = classloader_load(loader, classname);
 	if (!vmc) {
 		fprintf(stderr, "error: %s: could not load\n", classname);
@@ -973,7 +1080,7 @@ struct gnu_classpath_config gnu_classpath_configs[] = {
 	},
 };
 
-static void gnu_classpath_autodiscovery(void)
+static bool gnu_classpath_autodiscovery(void)
 {
 	for (unsigned int i = 0; i < ARRAY_SIZE(gnu_classpath_configs); i++) {
 		struct gnu_classpath_config *config = &gnu_classpath_configs[i];
@@ -981,9 +1088,49 @@ static void gnu_classpath_autodiscovery(void)
 		if (try_to_add_zip_to_classpath(config->glibj) < 0)
 			continue;
 
+		system_property_append_path("java.boot.class.path", config->glibj);
 		system_property_append_path("java.library.path", config->lib);
-		break;
+
+		return true;
 	}
+
+	return false;
+}
+
+static bool init_classpath(void)
+{
+	if (!system_property_get("java.boot.class.path")) {
+		if (!gnu_classpath_autodiscovery())
+			return false;
+	}
+
+	const char *boot_cp = system_property_get("java.boot.class.path");
+	add_system_property_const("sun.boot.class.path", boot_cp);
+
+	/* Search $CLASSPATH last. */
+	char *classpath = getenv("CLASSPATH");
+	if (classpath)
+		system_property_append_path("java.class.path", classpath);
+
+	if (!find_system_property("java.class.path")) {
+		const char *cwd = system_property_get("user.dir");
+		system_property_append_path("java.class.path", cwd);
+	}
+
+	return true;
+}
+
+static void print_proc_maps(void)
+{
+	char cmd[32];
+	pid_t self;
+
+	self = getpid();
+
+	snprintf(cmd, sizeof(cmd), "cat /proc/%d/maps", self);
+
+	if (system(cmd) != 0)
+		die("system");
 }
 
 int
@@ -1010,13 +1157,21 @@ main(int argc, char *argv[])
 
 	parse_options(argc, argv);
 
+	if (dump_maps)
+		print_proc_maps();
+
+	gc_init();
+	init_exec_env();
+	vm_reference_init();
+
 	classloader_init();
 
 	init_vm_objects();
-	init_vm_monitors();
 
 	jit_text_init();
 
+	gdb_init();
+
 	if (perf_enabled)
 		perf_map_open();
 
@@ -1026,18 +1181,13 @@ main(int argc, char *argv[])
 
 	jit_init_natives();
 
-	gc_init();
 	static_fixup_init();
 	vm_jni_init();
 
-	gnu_classpath_autodiscovery();
-
-	/* Search $CLASSPATH last. */
-	char *classpath = getenv("CLASSPATH");
-	if (classpath)
-		classloader_add_to_classpath(classpath);
-
-	init_configurable_system_properties();
+	if (!init_classpath()) {
+		fprintf(stderr, "Unable to locate GNU Classpath. Please specify 'java.boot.class.path' and 'java.library.path' manually.\n");
+		exit(EXIT_FAILURE);
+	}
 
 	if (preload_vm_classes()) {
 		NOT_IMPLEMENTED;
