@@ -36,10 +36,12 @@
 #include "vm/class.h"
 #include "vm/classloader.h"
 #include "vm/die.h"
+#include "vm/errors.h"
 #include "vm/jni.h"
 #include "vm/method.h"
 #include "vm/object.h"
 #include "vm/preload.h"
+#include "vm/reflection.h"
 #include "vm/stack-trace.h"
 
 #define check_null(x)							\
@@ -97,7 +99,7 @@ vm_jni_common_get_field_id(jclass clazz, const char *name, const char *sig)
 	if (exception_occurred())
 		return NULL;
 
-	fb = vm_class_get_field(class, name, sig);
+	fb = vm_class_get_field_recursive(class, name, sig);
 	if (!fb) {
 		signal_new_exception(vm_java_lang_NoSuchFieldError, NULL);
 		return NULL;
@@ -162,6 +164,14 @@ struct java_vm *vm_jni_get_current_java_vm(void)
 	return &vm_jni_default_java_vm;
 }
 
+static jobject vm_alloc_object(struct vm_jni_env *env, jclass clazz)
+{
+	struct vm_class *class = vm_class_get_class_from_class_object(clazz);
+	check_null(class);
+
+	return vm_object_alloc(class);
+}
+
 static jclass
 vm_jni_find_class(struct vm_jni_env *env, const char *name)
 {
@@ -516,7 +526,7 @@ static jint vm_jni_monitor_enter(struct vm_jni_env *env, jobject obj)
 {
 	enter_vm_from_jni();
 
-	int err = vm_monitor_lock(&obj->monitor);
+	int err = vm_object_lock(obj);
 
 	if (exception_occurred())
 		clear_exception();
@@ -528,7 +538,7 @@ static jint vm_jni_monitor_exit(struct vm_jni_env *env, jobject obj)
 {
 	enter_vm_from_jni();
 
-	int err = vm_monitor_unlock(&obj->monitor);
+	int err = vm_object_unlock(obj);
 
 	if (exception_occurred())
 		clear_exception();
@@ -604,11 +614,11 @@ vm_jni_get_ ## type ## _array_elements(struct vm_jni_env *env,		\
 	    != vm_ ## type ## _class)					\
 		return NULL;						\
 									\
-	result = malloc(sizeof(j ## type) * array->array_length);	\
+	result = malloc(sizeof(j ## type) * vm_array_length(array));	\
 	if (!result)							\
 		return NULL;						\
 									\
-	for (long i = 0; i < array->array_length; i++)			\
+	for (long i = 0; i < vm_array_length(array); i++)		\
 		result[i] = array_get_field_##type(array, i);		\
 									\
 	if (is_copy)							\
@@ -640,11 +650,11 @@ vm_jni_release_ ## type ## _array_elements(struct vm_jni_env *env,	\
 		return;							\
 									\
 	if (mode == 0 || mode == JNI_COMMIT) { /* copy back */		\
-		for (long i = 0; i < array->array_length; i++)		\
-			array_set_field_ ## type(array, i, elems[i]); \
+		for (long i = 0; i < vm_array_length(array); i++)	\
+			array_set_field_ ## type(array, i, elems[i]);	\
 	}								\
 									\
-	if (mode == 0 || mode == JNI_ABORT) /* free buffer */		\
+	if (mode == JNI_ABORT) /* free buffer */			\
 		free(elems);						\
 }
 
@@ -657,21 +667,220 @@ DECLARE_RELEASE_XXX_ARRAY_ELEMENTS(long);
 DECLARE_RELEASE_XXX_ARRAY_ELEMENTS(short);
 DECLARE_RELEASE_XXX_ARRAY_ELEMENTS(boolean);
 
+#define DECLARE_GET_XXX_ARRAY_CRITICAL(type)				\
+static void *								\
+get_ ## type ## _array_critical (jobject array, jboolean *is_copy)	\
+{									\
+	j ## type *result;						\
+									\
+	result = malloc(sizeof(j ## type) * vm_array_length(array));	\
+	if (!result)							\
+		return NULL;						\
+									\
+	for (long i = 0; i < vm_array_length(array); i++)		\
+		result[i] = array_get_field_##type(array, i);		\
+									\
+	if (is_copy)							\
+		*is_copy = JNI_TRUE;					\
+									\
+	return result;							\
+}
+
+#define DECLARE_RELEASE_XXX_ARRAY_CRITICAL(type)			\
+static void								\
+release_ ## type ## _array_critical (jobject array, j ## type *elems, jint mode) \
+{									\
+	if (mode == 0 || mode == JNI_COMMIT) { /* copy back */		\
+		for (long i = 0; i < vm_array_length(array); i++)	\
+			array_set_field_ ## type(array, i, elems[i]);	\
+	}								\
+									\
+	if (mode == JNI_ABORT) /* free buffer */			\
+		free(elems);						\
+}
+
+DECLARE_GET_XXX_ARRAY_CRITICAL(byte);
+DECLARE_GET_XXX_ARRAY_CRITICAL(char);
+DECLARE_GET_XXX_ARRAY_CRITICAL(double);
+DECLARE_GET_XXX_ARRAY_CRITICAL(float);
+DECLARE_GET_XXX_ARRAY_CRITICAL(int);
+DECLARE_GET_XXX_ARRAY_CRITICAL(long);
+DECLARE_GET_XXX_ARRAY_CRITICAL(short);
+DECLARE_GET_XXX_ARRAY_CRITICAL(boolean);
+
+DECLARE_RELEASE_XXX_ARRAY_CRITICAL(byte);
+DECLARE_RELEASE_XXX_ARRAY_CRITICAL(char);
+DECLARE_RELEASE_XXX_ARRAY_CRITICAL(double);
+DECLARE_RELEASE_XXX_ARRAY_CRITICAL(float);
+DECLARE_RELEASE_XXX_ARRAY_CRITICAL(int);
+DECLARE_RELEASE_XXX_ARRAY_CRITICAL(long);
+DECLARE_RELEASE_XXX_ARRAY_CRITICAL(short);
+DECLARE_RELEASE_XXX_ARRAY_CRITICAL(boolean);
+
+typedef void *get_array_critical_fn(jobject array, jboolean *is_copy);
+typedef void release_array_critical_fn(jobject array, void *carray, jint mode);
+
+get_array_critical_fn *get_array_critical_fns[] = {
+	[J_BYTE]	= get_byte_array_critical,
+	[J_CHAR]	= get_char_array_critical,
+	[J_DOUBLE]	= get_double_array_critical,
+	[J_FLOAT]	= get_float_array_critical,
+	[J_INT]		= get_int_array_critical,
+	[J_LONG]	= get_long_array_critical,
+	[J_SHORT]	= get_short_array_critical,
+	[J_BOOLEAN]	= get_boolean_array_critical
+};
+
+release_array_critical_fn *release_array_critical_fns[] = {
+	[J_BYTE]	= (release_array_critical_fn*) release_byte_array_critical,
+	[J_CHAR]	= (release_array_critical_fn*) release_char_array_critical,
+	[J_DOUBLE]	= (release_array_critical_fn*) release_double_array_critical,
+	[J_FLOAT]	= (release_array_critical_fn*) release_float_array_critical,
+	[J_INT]		= (release_array_critical_fn*) release_int_array_critical,
+	[J_LONG]	= (release_array_critical_fn*) release_long_array_critical,
+	[J_SHORT]	= (release_array_critical_fn*) release_short_array_critical,
+	[J_BOOLEAN]	= (release_array_critical_fn*) release_boolean_array_critical
+};
+
 static void *
-vm_jni_get_direct_buffer_address(struct vm_jni_env *env, jobject buf)
+vm_jni_get_primitive_array_critical(struct vm_jni_env *env,
+				    jobject array,
+				    jboolean *is_copy)
 {
 	enter_vm_from_jni();
 
-	/* We can't return direct buffer because we use machine word size
-	   elements for arrays. */
+	if (!vm_class_is_array_class(array->class))
+		return NULL;
+
+	struct vm_class *elem_class
+		= vm_class_get_array_element_class(array->class);
+
+	if (!elem_class)
+		return rethrow_exception();
+
+	if (!vm_class_is_primitive_class(elem_class))
+		return NULL;
+
+	enum vm_type elem_type
+		= vm_class_get_storage_vmtype(elem_class);
+
+	get_array_critical_fn *get_array_critical
+		= get_array_critical_fns[elem_type];
+
+	assert(get_array_critical);
+	return get_array_critical(array, is_copy);
+}
+
+
+static void
+vm_jni_release_primitive_array_critical(struct vm_jni_env *env,
+					jobject array,
+					void *carray,
+					jint mode)
+{
+	enter_vm_from_jni();
+
+	if (!vm_class_is_array_class(array->class))
+		return;
+
+	struct vm_class *elem_class
+		= vm_class_get_array_element_class(array->class);
+
+	if (!elem_class)
+		return;
+
+	if (!vm_class_is_primitive_class(elem_class))
+		return;
+
+	enum vm_type elem_type
+		= vm_class_get_storage_vmtype(elem_class);
+
+	release_array_critical_fn *release_array_critical
+		= release_array_critical_fns[elem_type];
+
+	assert(release_array_critical);
+	release_array_critical(array, carray, mode);
+}
+
+static jobject JNI_NewDirectByteBuffer(struct vm_jni_env *env, void *address, jlong capacity)
+{
+	enter_vm_from_jni();
+
+	error("%s is not supported", __func__);
+
 	return NULL;
 }
 
-/*
- * TODO: When these functions are used to call private methods and
- * constructors, the method ID must be derived from the real class of
- * obj, not from one of its superclasses.
+static void *JNI_GetDirectBufferAddress(struct vm_jni_env *env, jobject buf)
+{
+	struct vm_object *address;
+	void *data;
+
+	enter_vm_from_jni();
+
+	if (buf == NULL)
+		return NULL;
+
+	address		= field_get_object(buf, vm_java_nio_Buffer_address);
+	if (!address)
+		return NULL;
+
+#ifdef CONFIG_32_BIT
+	data		= (void *) field_get_int(buf, vm_gnu_classpath_PointerNN_data);
+#else
+	data		= (void *) field_get_long(buf, vm_gnu_classpath_PointerNN_data);
+#endif
+
+	return data;
+}
+
+static jlong JNI_GetDirectBufferCapacity(struct vm_jni_env *env, jobject buf)
+{
+	enter_vm_from_jni();
+
+	error("%s is not supported", __func__);
+
+	return 0;
+}
+
+/**
+ * When given method is a private methods or constructor, the method
+ * must be derived from the real class of object, not from one of its
+ * superclasses. It's used by CallXXXMethod() function family.
  */
+static int transform_method_for_call(jobject object, jmethodID *method_p)
+{
+	struct vm_method *vmm = *method_p;
+
+	if (!vm_method_is_private(vmm) && !vm_method_is_constructor(vmm))
+		return 0;
+
+	if (!object) {
+		throw_npe();
+		return -1;
+	}
+
+	struct vm_class *actual_class = object->class;
+
+	if (!vm_class_is_assignable_from(vmm->class, actual_class)) {
+		signal_new_exception(vm_java_lang_Error,
+				     "Object not assignable to %s",
+				     vmm->class->name);
+		return -1;
+	}
+
+	vmm = vm_class_get_method_recursive(actual_class, vmm->name, vmm->type);
+	if (!vmm) {
+		signal_new_exception(vm_java_lang_NoSuchMethodError, "%s%s in %s",
+				     (*method_p)->name, (*method_p)->type,
+				     actual_class->name);
+		return -1;
+	}
+
+	*method_p = vmm;
+	return 0;
+}
+
 static void vm_jni_call_void_method(struct vm_jni_env *env, jobject this,
 				    jmethodID methodID, ...)
 {
@@ -679,21 +888,33 @@ static void vm_jni_call_void_method(struct vm_jni_env *env, jobject this,
 
 	enter_vm_from_jni();
 
+	if (transform_method_for_call(this, &methodID))
+		return;
+
 	va_start(args, methodID);
 	vm_call_method_this_v(methodID, this, args, NULL);
 	va_end(args);
 }
 
-/*
- * TODO: When these functions are used to call private methods and
- * constructors, the method ID must be derived from the real class of
- * obj, not from one of its superclasses.
- */
-static void vm_jni_call_void_method_a(struct vm_jni_env *env, jobject this,
-				    jmethodID methodID, uint64_t *args)
+static void vm_jni_call_void_method_v(struct vm_jni_env *env, jobject this,
+				      jmethodID methodID, va_list args)
 {
 	enter_vm_from_jni();
 
+	if (transform_method_for_call(this, &methodID))
+		return;
+
+	vm_call_method_this_v(methodID, this, args, NULL);
+}
+
+static void vm_jni_call_void_method_a(struct vm_jni_env *env, jobject this,
+				      jmethodID methodID, uint64_t *args)
+{
+	enter_vm_from_jni();
+
+	if (transform_method_for_call(this, &methodID))
+		return;
+
 	unsigned long packed_args[methodID->args_count];
 
 	packed_args[0] = (unsigned long) this;
@@ -702,12 +923,6 @@ static void vm_jni_call_void_method_a(struct vm_jni_env *env, jobject this,
 	vm_call_method_this_a(methodID, this, packed_args, NULL);
 }
 
-
-/*
- * TODO: When these functions are used to call private methods and
- * constructors, the method ID must be derived from the real class of
- * obj, not from one of its superclasses.
- */
 #define DECLARE_CALL_XXX_METHOD(type, symbol)				\
 	static j ## type vm_jni_call_ ## type ## _method		\
 	(struct vm_jni_env *env, jobject this, jmethodID methodID, ...)	\
@@ -717,6 +932,9 @@ static void vm_jni_call_void_method_a(struct vm_jni_env *env, jobject this,
 									\
 		enter_vm_from_jni();					\
 									\
+		if (transform_method_for_call(this, &methodID))		\
+			return 0;					\
+									\
 		va_start(args, methodID);				\
 		vm_call_method_this_v(methodID, this, args, &result);	\
 		va_end(args);						\
@@ -724,6 +942,22 @@ static void vm_jni_call_void_method_a(struct vm_jni_env *env, jobject this,
 		return result.symbol;					\
 	}
 
+#define DECLARE_CALL_XXX_METHOD_V(type, symbol)				\
+        static j ## type vm_jni_call_ ## type ## _method_v		\
+        (struct vm_jni_env *env, jobject this, jmethodID methodID, va_list args) \
+	{								\
+		union jvalue result;                                    \
+									\
+		enter_vm_from_jni();                                    \
+									\
+		if (transform_method_for_call(this, &methodID))         \
+			return 0;                                       \
+									\
+		vm_call_method_this_v(methodID, this, args, &result);   \
+									\
+		return result.symbol;                                   \
+	}
+
 DECLARE_CALL_XXX_METHOD(boolean, z);
 DECLARE_CALL_XXX_METHOD(byte, b);
 DECLARE_CALL_XXX_METHOD(char, c);
@@ -734,6 +968,114 @@ DECLARE_CALL_XXX_METHOD(float, f);
 DECLARE_CALL_XXX_METHOD(double, d);
 DECLARE_CALL_XXX_METHOD(object, l);
 
+DECLARE_CALL_XXX_METHOD_V(boolean, z);
+DECLARE_CALL_XXX_METHOD_V(byte, b);
+DECLARE_CALL_XXX_METHOD_V(char, c);
+DECLARE_CALL_XXX_METHOD_V(short, s);
+DECLARE_CALL_XXX_METHOD_V(int, i);
+DECLARE_CALL_XXX_METHOD_V(long, j);
+DECLARE_CALL_XXX_METHOD_V(float, f);
+DECLARE_CALL_XXX_METHOD_V(double, d);
+DECLARE_CALL_XXX_METHOD_V(object, l);
+
+static int transform_method_for_nonvirtual_call(jobject this, jclass clazz,
+						jmethodID *method_p)
+{
+	struct vm_method *vmm = *method_p;
+
+	if (!this || !clazz || !vmm) {
+		throw_npe();
+		return -1;
+	}
+
+	struct vm_class *vmc =
+		vm_class_get_class_from_class_object(clazz);
+
+	if (!vmc) {
+		throw_npe();
+		return -1;
+	}
+
+	if (!vm_class_is_assignable_from(vmm->class, this->class)) {
+		signal_new_exception(vm_java_lang_Error,
+				     "Object is not assignable to %s",
+				     vmm->class->name);
+		return -1; /* rethrow */
+	}
+
+	vmm = vm_class_get_method_recursive(vmc, vmm->name, vmm->type);
+	if (!vmm) {
+		signal_new_exception(vm_java_lang_NoSuchMethodError, "%s%s",
+				     (*method_p)->name, (*method_p)->type);
+		return -1; /* rethrow */
+	}
+
+	*method_p = vmm;
+	return 0;
+}
+
+static void vm_jni_call_nonvirtual_void_method(struct vm_jni_env *env,
+	jobject this, jclass clazz, jmethodID methodID, ...)
+{
+	va_list args;
+
+	enter_vm_from_jni();
+
+	if (transform_method_for_nonvirtual_call(this, clazz, &methodID))
+		return; /* rethrow */
+
+	va_start(args, methodID);
+	vm_call_method_this_v(methodID, this, args, NULL);
+	va_end(args);
+}
+
+static void vm_jni_call_nonvirtual_void_method_a(struct vm_jni_env *env,
+	jobject this, jclass clazz, jmethodID methodID, uint64_t *args)
+{
+	enter_vm_from_jni();
+
+	if (transform_method_for_nonvirtual_call(this, clazz, &methodID))
+		return; /* rethrow */
+
+	unsigned long packed_args[methodID->args_count];
+
+	packed_args[0] = (unsigned long) this;
+	pack_args(methodID, packed_args + 1, args);
+
+	vm_call_method_this_a(methodID, this, packed_args, NULL);
+}
+
+#define DECLARE_CALL_NONVIRTUAL_XXX_METHOD(type, symbol)		\
+	static j ## type vm_jni_call_nonvirtual_ ## type ## _method	\
+	(struct vm_jni_env *env, jobject this, jclass clazz,		\
+	 jmethodID methodID, ...)					\
+	{								\
+		va_list args;						\
+		union jvalue result;					\
+									\
+		enter_vm_from_jni();					\
+									\
+		if (transform_method_for_nonvirtual_call(this, clazz,	\
+							 &methodID))	\
+			return 0; /* rethrow */				\
+									\
+									\
+		va_start(args, methodID);				\
+		vm_call_method_this_v(methodID, this, args, &result);	\
+		va_end(args);						\
+									\
+		return result.symbol;					\
+	}
+
+DECLARE_CALL_NONVIRTUAL_XXX_METHOD(boolean, z);
+DECLARE_CALL_NONVIRTUAL_XXX_METHOD(byte, b);
+DECLARE_CALL_NONVIRTUAL_XXX_METHOD(char, c);
+DECLARE_CALL_NONVIRTUAL_XXX_METHOD(int, i);
+DECLARE_CALL_NONVIRTUAL_XXX_METHOD(long, j);
+DECLARE_CALL_NONVIRTUAL_XXX_METHOD(float, f);
+DECLARE_CALL_NONVIRTUAL_XXX_METHOD(double, d);
+DECLARE_CALL_NONVIRTUAL_XXX_METHOD(object, l);
+
 static jfieldID
 vm_jni_get_static_field_id(struct vm_jni_env *env, jclass clazz,
 			   const char *name, const char *sig)
@@ -806,16 +1148,10 @@ vm_jni_new_object_array(struct vm_jni_env *env, jsize size,
 	if (!vmc)
 		return NULL;
 
-	vmc = vm_class_get_array_class(vmc);
-	if (!vmc)
+	array = vm_object_alloc_array_of(vmc, size);
+	if (!array)
 		return NULL;
 
-	array = vm_object_alloc_array(vmc, size);
-	if (!array) {
-		NOT_IMPLEMENTED;
-		return NULL;
-	}
-
 	while (size)
 		array_set_field_object(array, --size, initial_element);
 
@@ -832,7 +1168,7 @@ static jobject vm_jni_get_object_array_element(struct vm_jni_env *env,
 		signal_new_exception(vm_java_lang_NullPointerException, NULL);
 		return NULL;
 	}
-	if (index >= array->array_length) {
+	if (index >= vm_array_length(array)) {
 		signal_new_exception(vm_java_lang_ArrayIndexOutOfBoundsException, NULL);
 		return NULL;
 	}
@@ -851,7 +1187,7 @@ static void vm_jni_set_object_array_element(struct vm_jni_env *env,
 		signal_new_exception(vm_java_lang_NullPointerException, NULL);
 		return;
 	}
-	if (index >= array->array_length) {
+	if (index >= vm_array_length(array)) {
 		signal_new_exception(vm_java_lang_ArrayIndexOutOfBoundsException, NULL);
 		return;
 	}
@@ -868,7 +1204,7 @@ static jsize vm_jni_get_array_length(struct vm_jni_env *env, jarray array)
 		return 0;
 	}
 
-	return array->array_length;
+	return vm_array_length(array);
 }
 
 #define DECLARE_NEW_XXX_ARRAY(type, arr_type)				\
@@ -929,7 +1265,7 @@ static void								\
 	    != vm_ ## type ## _class)					\
 		return;							\
 									\
-	if (start < 0 || len < 0 || start + len > array->array_length) { \
+	if (start < 0 || len < 0 || start + len > vm_array_length(array)) { \
 		signal_new_exception(vm_java_lang_ArrayIndexOutOfBoundsException, \
 				     NULL);				\
 		return;							\
@@ -963,7 +1299,7 @@ static void								\
 	    != vm_ ## type ## _class)					\
 		return;							\
 									\
-	if (start < 0 || len < 0 || start + len > array->array_length) { \
+	if (start < 0 || len < 0 || start + len > vm_array_length(array)) { \
 		signal_new_exception(vm_java_lang_ArrayIndexOutOfBoundsException, \
 				     NULL);				\
 		return;							\
@@ -973,6 +1309,12 @@ static void								\
 		array_set_field_##type(array, start + i, buf[i]);	\
 }
 
+static jmethodID vm_jni_from_reflected_method(struct vm_exec_env *env,
+					      jobject method)
+{
+	return vm_object_to_vm_method(method);
+}
+
 DECLARE_SET_XXX_ARRAY_REGION(byte);
 DECLARE_SET_XXX_ARRAY_REGION(char);
 DECLARE_SET_XXX_ARRAY_REGION(double);
@@ -982,9 +1324,36 @@ DECLARE_SET_XXX_ARRAY_REGION(long);
 DECLARE_SET_XXX_ARRAY_REGION(short);
 DECLARE_SET_XXX_ARRAY_REGION(boolean);
 
+static jboolean
+vm_jni_is_same_object(struct vm_exec_env *env, jobject o1, jobject o2)
+{
+	enter_vm_from_jni();
+
+	if (o1 == o2)
+		return JNI_TRUE;
+
+	return JNI_FALSE;
+}
+
+static jint vm_jni_get_string_length(struct vm_exec_env *env, jobject string)
+{
+	union jvalue result;
+
+	enter_vm_from_jni();
+
+	if (string->class != vm_java_lang_String) { /* String is a final */
+		signal_new_exception(vm_java_lang_IllegalArgumentException, NULL);
+		return 0; /* rethrow */
+	}
+
+	vm_call_method(vm_java_lang_String_length, string, &result);
+
+	return result.i;
+}
+
 /*
  * The JNI native interface table.
- * See: http://java.sun.com/j2se/1.4.2/docs/guide/jni/spec/functions.html
+ * See: http://download.oracle.com/javase/1.4.2/docs/guide/jni/spec/functions.html
  */
 void *vm_jni_native_interface[] = {
 	/* 0 */
@@ -997,7 +1366,7 @@ void *vm_jni_native_interface[] = {
 	/* 5 */
 	NULL, /* DefineClass */
 	vm_jni_find_class,
-	NULL,
+	vm_jni_from_reflected_method,
 	NULL,
 	NULL,
 
@@ -1020,12 +1389,12 @@ void *vm_jni_native_interface[] = {
 	vm_jni_new_global_ref,
 	vm_jni_delete_global_ref,
 	vm_jni_delete_local_ref,
-	NULL, /* IsSameObject */
+	vm_jni_is_same_object,
 
 	/* 25 */
 	NULL,
 	NULL,
-	NULL, /* AllocObject */
+	vm_alloc_object,
 	vm_jni_new_object,
 	NULL, /* NewObjectV */
 
@@ -1037,59 +1406,59 @@ void *vm_jni_native_interface[] = {
 	vm_jni_call_object_method,
 
 	/* 35 */
-	NULL, /* CallObjectMethodV */
+	vm_jni_call_object_method_v,
 	NULL, /* CallObjectMethodA */
 	vm_jni_call_boolean_method,
-	NULL, /* CallBooleanMethodV */
+	vm_jni_call_boolean_method_v,
 	NULL, /* CallBooleanMethodA */
 
 	/* 40 */
 	vm_jni_call_byte_method,
-	NULL, /* CallByteMethodV */
+	vm_jni_call_byte_method_v,
 	NULL, /* CallByteMethodA */
 	vm_jni_call_char_method,
-	NULL, /* CallCharMethodV */
+	vm_jni_call_char_method_v,
 
 	/* 45 */
 	NULL, /* CallCharMethodA */
 	vm_jni_call_short_method,
-	NULL, /* CallShortMethodV */
+	vm_jni_call_short_method_v,
 	NULL, /* CallShortMethodA */
 	vm_jni_call_int_method,
 
 	/* 50 */
-	NULL, /* CallIntMethodV */
+	vm_jni_call_int_method_v,
 	NULL, /* CallIntMethodA */
 	vm_jni_call_long_method,
-	NULL, /* CallLongMethodV */
+	vm_jni_call_long_method_v,
 	NULL, /* CallLongMethodA */
 
 	/* 55 */
 	vm_jni_call_float_method,
-	NULL, /* CallFloatMethodV */
+	vm_jni_call_float_method_v,
 	NULL, /* CallFloatMethodA */
 	vm_jni_call_double_method,
-	NULL, /* CallDoubleMethodV */
+	vm_jni_call_double_method_v,
 
 	/* 60 */
 	NULL, /* CallDoubleMethodA */
 	vm_jni_call_void_method,
-	NULL, /* CallVoidMethodV */
+	vm_jni_call_void_method_v,
 	vm_jni_call_void_method_a,
-	NULL, /* CallNonvirtualObjectMethod */
+	vm_jni_call_nonvirtual_object_method,
 
 	/* 65 */
 	NULL, /* CallNonvirtualObjectMethodV */
 	NULL, /* CallNonvirtualObjectMethodA */
-	NULL, /* CallNonvirtualBooleanMethod */
+	vm_jni_call_nonvirtual_boolean_method,
 	NULL, /* CallNonvirtualBooleanMethodV */
 	NULL, /* CallNonvirtualBooleanMethodA */
 
 	/* 70 */
-	NULL, /* CallNonvirtualByteMethod */
+	vm_jni_call_nonvirtual_byte_method,
 	NULL, /* CallNonvirtualByteMethodV */
 	NULL, /* CallNonvirtualByteMethodA */
-	NULL, /* CallNonvirtualCharMethod */
+	vm_jni_call_nonvirtual_char_method,
 	NULL, /* CallNonvirtualCharMethodV */
 
 	/* 75 */
@@ -1097,27 +1466,27 @@ void *vm_jni_native_interface[] = {
 	NULL, /* CallNonvirtualShortMethod */
 	NULL, /* CallNonvirtualShortMethodV */
 	NULL, /* CallNonvirtualShortMethodA */
-	NULL, /* CallNonvirtualIntMethod */
+	vm_jni_call_nonvirtual_int_method,
 
 	/* 80 */
 	NULL, /* CallNonvirtualIntMethodV */
 	NULL, /* CallNonvirtualIntMethodA */
-	NULL, /* CallNonvirtualLongMethod */
+	vm_jni_call_nonvirtual_long_method,
 	NULL, /* CallNonvirtualLongMethodV */
 	NULL, /* CallNonvirtualLongMethodA */
 
 	/* 85 */
-	NULL, /* CallNonvirtualFloatMethod */
+	vm_jni_call_nonvirtual_float_method,
 	NULL, /* CallNonvirtualFloatMethodV */
 	NULL, /* CallNonvirtualFloatMethodA */
-	NULL, /* CallNonvirtualDoubleMethod */
+	vm_jni_call_nonvirtual_double_method,
 	NULL, /* CallNonvirtualDoubleMethodV */
 
 	/* 90 */
 	NULL, /* CallNonvirtualDoubleMethodA */
-	NULL, /* CallNonvirtualVoidMethod */
+	vm_jni_call_nonvirtual_void_method,
 	NULL, /* CallNonvirtualVoidMethodV */
-	NULL, /* CallNonvirtualVoidMethodA */
+	vm_jni_call_nonvirtual_void_method_a,
 	vm_jni_get_field_id,
 
 	/* 95 */
@@ -1216,7 +1585,7 @@ void *vm_jni_native_interface[] = {
 	vm_jni_set_static_float_field,
 	vm_jni_set_static_double_field,
 	NULL, /* NewString */
-	NULL, /* GetStringLength */
+	vm_jni_get_string_length,
 
 	/* 165 */
 	NULL, /* GetStringChars */
@@ -1300,8 +1669,8 @@ void *vm_jni_native_interface[] = {
 	/* 220 */
 	NULL, /* GetStringRegion */
 	NULL, /* GetStringUTFRegion */
-	NULL, /* GetPrimitiveArrayCritical */
-	NULL, /* ReleasePrimitiveArrayCritical */
+	vm_jni_get_primitive_array_critical,
+	vm_jni_release_primitive_array_critical,
 	NULL, /* GetStringCritical */
 
 	/* 225 */
@@ -1312,11 +1681,11 @@ void *vm_jni_native_interface[] = {
 
 	/* JNI 1.4 functions */
 
-	NULL, /* NewDirectByteBuffer */
+	JNI_NewDirectByteBuffer,
 
 	/* 230 */
-	vm_jni_get_direct_buffer_address,
-	NULL, /* GetDirectBufferCapacity */
+	JNI_GetDirectBufferAddress,
+	JNI_GetDirectBufferCapacity,
 	NULL,
 	NULL,
 	NULL,
