@@ -1,7 +1,3 @@
-#include <string.h>
-#include <stdio.h>
-#include <stdlib.h>
-
 #include "cafebabe/attribute_array.h"
 #include "cafebabe/attribute_info.h"
 #include "cafebabe/line_number_table_attribute.h"
@@ -15,8 +11,14 @@
 #include "vm/method.h"
 #include "vm/natives.h"
 
-#include "jit/args.h"
+#include "jit/compilation-unit.h"
 #include "jit/cu-mapping.h"
+#include "jit/args.h"
+#include "jit/gdb.h"
+
+#include <stdlib.h>
+#include <string.h>
+#include <stdio.h>
 
 static void init_abstract_method(struct vm_method *vmm)
 {
@@ -28,6 +30,35 @@ static void init_abstract_method(struct vm_method *vmm)
 	vmm->line_number_table_attribute.line_number_table = NULL;
 }
 
+int vm_method_do_init(struct vm_method *vmm)
+{
+	if (parse_method_type(vmm)) {
+		warn("method type parsing failed for: %s", vmm->type);
+		return -1;
+	}
+
+	/*
+	 * XXX: Jam VM legacy? It seems that JamVM counts the number of
+	 * _32-bit_ arguments. This probably needs some fixing. What do we do
+	 * on x86_64?
+	 */
+	vmm->args_count = count_arguments(vmm);
+	if (vmm->args_count < 0)
+		return -1;
+
+	if (!vm_method_is_static(vmm))
+		++vmm->args_count;
+
+	if (vm_method_is_native(vmm) && vm_lookup_native(vmm->class->name, vmm->name)) {
+		vmm->flags |= VM_METHOD_FLAG_VM_NATIVE;
+	}
+
+	if (args_map_init(vmm))
+		return -1;
+
+	return 0;
+}
+
 int vm_method_init(struct vm_method *vmm,
 	struct vm_class *vmc, unsigned int method_index)
 {
@@ -38,65 +69,26 @@ int vm_method_init(struct vm_method *vmm,
 	vmm->class = vmc;
 	vmm->method_index = method_index;
 	vmm->method = method;
+	vmm->flags = 0;
 
 	const struct cafebabe_constant_info_utf8 *name;
-	if (cafebabe_class_constant_get_utf8(class, method->name_index,
-		&name))
-	{
-		NOT_IMPLEMENTED;
+	if (cafebabe_class_constant_get_utf8(class, method->name_index, &name))
 		return -1;
-	}
 
 	vmm->name = strndup((char *) name->bytes, name->length);
-	if (!vmm->name) {
-		NOT_IMPLEMENTED;
+	if (!vmm->name)
 		return -1;
-	}
 
 	const struct cafebabe_constant_info_utf8 *type;
-	if (cafebabe_class_constant_get_utf8(class, method->descriptor_index,
-		&type))
-	{
-		NOT_IMPLEMENTED;
-		return -1;
-	}
+	if (cafebabe_class_constant_get_utf8(class, method->descriptor_index, &type))
+		goto error_free_name;
 
 	vmm->type = strndup((char *) type->bytes, type->length);
-	if (!vmm->type) {
-		NOT_IMPLEMENTED;
-		return -1;
-	}
+	if (!vmm->type)
+		goto error_free_name;
 
-	if (parse_method_type(vmm)) {
-		warn("method type parsing failed for: %s", vmm->type);
-		return -1;
-	}
-
-	/*
-	 * XXX: Jam VM legacy? It seems that JamVM counts the number of
-	 * _32-bit_ arguments. This probably needs some fixing. What do we do
-	 * on x86_64?
-	 */
-	vmm->args_count = count_arguments(vmm);
-	if (vmm->args_count < 0) {
-		NOT_IMPLEMENTED;
-		return -1;
-	}
-
-	if (!vm_method_is_static(vmm))
-		++vmm->args_count;
-
-	vmm->is_vm_native = false;
-
-	if (vm_method_is_native(vmm)) {
-		vmm->is_vm_native =
-			vm_lookup_native(vmm->class->name, vmm->name);
-	}
-
-	if (args_map_init(vmm)) {
-		abort();
-		return -1;
-	}
+	if (vm_method_do_init(vmm))
+		goto error_free_type;
 
 	/*
 	 * Note: We can return here because the rest of the function deals
@@ -108,23 +100,13 @@ int vm_method_init(struct vm_method *vmm,
 	}
 
 	unsigned int code_index = 0;
-	if (cafebabe_attribute_array_get(&method->attributes,
-		"Code", class, &code_index))
-	{
-		NOT_IMPLEMENTED;
-		return -1;
-	}
+	if (cafebabe_attribute_array_get(&method->attributes, "Code", class, &code_index))
+		goto error_free_type;
 
-	{
-		/* There must be only one "Code" attribute for the method! */
-		unsigned int code_index2 = code_index + 1;
-		if (!cafebabe_attribute_array_get(&method->attributes,
-			"Code", class, &code_index2))
-		{
-			NOT_IMPLEMENTED;
-			return -1;
-		}
-	}
+	/* There must be only one "Code" attribute for the method! */
+	unsigned int code_index2 = code_index + 1;
+	if (!cafebabe_attribute_array_get(&method->attributes, "Code", class, &code_index2))
+		goto error_free_type;
 
 	const struct cafebabe_attribute_info *attribute
 		= &method->attributes.array[code_index];
@@ -133,22 +115,25 @@ int vm_method_init(struct vm_method *vmm,
 	cafebabe_stream_open_buffer(&stream,
 		attribute->info, attribute->attribute_length);
 
-	if (cafebabe_code_attribute_init(&vmm->code_attribute, &stream)) {
-		NOT_IMPLEMENTED;
-		return -1;
-	}
+	if (cafebabe_code_attribute_init(&vmm->code_attribute, &stream))
+		goto error_free_type;
 
 	cafebabe_stream_close_buffer(&stream);
 
-	if (cafebabe_read_line_number_table_attribute(class,
-		&vmm->code_attribute.attributes,
-		&vmm->line_number_table_attribute))
-	{
-		NOT_IMPLEMENTED;
-		return -1;
-	}
+	if (cafebabe_read_exceptions_attribute(class, &method->attributes, &vmm->exceptions_attribute))
+		goto error_free_type;
+
+	if (cafebabe_read_line_number_table_attribute(class, &vmm->code_attribute.attributes, &vmm->line_number_table_attribute))
+		goto error_free_type;
 
 	return 0;
+
+error_free_type:
+	free(vmm->type);
+error_free_name:
+	free(vmm->name);
+
+	return -1;
 }
 
 int vm_method_init_from_interface(struct vm_method *vmm, struct vm_class *vmc,
@@ -167,7 +152,7 @@ int vm_method_init_from_interface(struct vm_method *vmm, struct vm_class *vmc,
 	vmm->type = interface_method->type;
 
 	vmm->args_count = interface_method->args_count;
-	vmm->is_vm_native = false;
+	vmm->flags = 0;
 
 	if (parse_method_type(vmm)) {
 		warn("method type parsing failed for: %s", vmm->type);
@@ -184,31 +169,38 @@ int vm_method_prepare_jit(struct vm_method *vmm)
 	struct compilation_unit *cu;
 
 	cu = compilation_unit_alloc(vmm);
-	if (!cu) {
-		NOT_IMPLEMENTED;
+	if (!cu)
 		return -1;
-	}
 
 	vmm->compilation_unit = cu;
 
+	if (method_matches_regex(vmm))
+		vmm->flags |= VM_METHOD_FLAG_TRACE;
+
+	if (method_matches_gate_regex(vmm))
+		vmm->flags |= VM_METHOD_FLAG_TRACE_GATE;
+
 	/*
 	 * VM native methods are linked on initialization.
 	 */
 	if (vm_method_is_vm_native(vmm)) {
 		cu->native_ptr = vm_lookup_native(vmm->class->name, vmm->name);
-		cu->is_compiled = true;
+		compile_lock_set_status(&cu->compile_lock, STATUS_COMPILED_OK);
 
-		if (add_cu_mapping((unsigned long)cu->native_ptr, cu)) {
-			NOT_IMPLEMENTED;
-			return -1;
-		}
+		if (add_cu_mapping((unsigned long)cu->native_ptr, cu))
+			goto error_free_cu;
 	}
 
 	vmm->trampoline = build_jit_trampoline(cu);
-	if (!vmm->trampoline) {
-		NOT_IMPLEMENTED;
-		return -1;
-	}
+	if (!vmm->trampoline)
+		goto error_free_cu;
+
+	gdb_register_trampoline(vmm);
 
 	return 0;
+
+error_free_cu:
+	free_compilation_unit(cu);
+
+	return -1;
 }
