@@ -13,6 +13,7 @@
 #include "vm/utf8.h"
 #include "vm/die.h"
 #include "vm/gc.h"
+#include "vm/reference.h"
 
 #include "lib/string.h"
 
@@ -41,9 +42,19 @@ int init_vm_objects(void)
 	return 0;
 }
 
-static void vm_object_finalizer(struct vm_object *obj, void *param)
+void vm_object_finalizer(struct vm_object *obj)
+{
+	vm_reference_collect_for_object(obj);
+
+	if (vm_object_is_instance_of(obj, vm_java_lang_ref_Reference))
+		vm_reference_collect(obj);
+
+	if (obj->class == vm_java_lang_VMThread)
+		vm_thread_collect_vmthread(obj);
+}
+
+static void vm_object_init_common(struct vm_object *object)
 {
-	// TODO: implement
 }
 
 struct vm_object *vm_object_alloc(struct vm_class *class)
@@ -57,19 +68,15 @@ struct vm_object *vm_object_alloc(struct vm_class *class)
 	if (!res)
 		return throw_oom_error();
 
-	gc_register_finalizer(res, vm_object_finalizer, NULL);
+	vm_object_init_common(res);
 
 	res->class = class;
-
-	if (vm_monitor_init(&res->monitor))
-		return throw_internal_error();
-
 	return res;
 }
 
 struct vm_object *vm_object_alloc_primitive_array(int type, int count)
 {
-	struct vm_object *res;
+	struct vm_array *res;
 	int vm_type;
 
 	vm_type = bytecode_type_to_vmtype(type);
@@ -79,56 +86,53 @@ struct vm_object *vm_object_alloc_primitive_array(int type, int count)
 	if (!res)
 		return throw_oom_error();
 
-	gc_register_finalizer(res, vm_object_finalizer, NULL);
+	vm_object_init_common(&res->object);
 
 	switch (type) {
 	case T_BOOLEAN:
-		res->class = classloader_load(NULL, "[Z");
+		res->object.class = classloader_load(NULL, "[Z");
 		break;
 	case T_CHAR:
-		res->class = classloader_load(NULL, "[C");
+		res->object.class = classloader_load(NULL, "[C");
 		break;
 	case T_FLOAT:
-		res->class = classloader_load(NULL, "[F");
+		res->object.class = classloader_load(NULL, "[F");
 		break;
 	case T_DOUBLE:
-		res->class = classloader_load(NULL, "[D");
+		res->object.class = classloader_load(NULL, "[D");
 		break;
 	case T_BYTE:
-		res->class = classloader_load(NULL, "[B");
+		res->object.class = classloader_load(NULL, "[B");
 		break;
 	case T_SHORT:
-		res->class = classloader_load(NULL, "[S");
+		res->object.class = classloader_load(NULL, "[S");
 		break;
 	case T_INT:
-		res->class = classloader_load(NULL, "[I");
+		res->object.class = classloader_load(NULL, "[I");
 		break;
 	case T_LONG:
-		res->class = classloader_load(NULL, "[J");
+		res->object.class = classloader_load(NULL, "[J");
 		break;
 	default:
 		return throw_internal_error();
 	}
 
-	if (!res->class)
+	if (!res->object.class)
 		return throw_internal_error();
 
-	if (vm_class_ensure_init(res->class))
+	if (vm_class_ensure_init(res->object.class))
 		return throw_internal_error();
 
 	res->array_length = count;
 
-	if (vm_monitor_init(&res->monitor))
-		return throw_internal_error();
-
-	return res;
+	return &res->object;
 }
 
 struct vm_object *
 vm_object_alloc_multi_array(struct vm_class *class, int nr_dimensions, int *counts)
 {
 	struct vm_class *elem_class;
-	struct vm_object *res;
+	struct vm_array *res;
 	int elem_size;
 
 	assert(nr_dimensions > 0);
@@ -143,28 +147,24 @@ vm_object_alloc_multi_array(struct vm_class *class, int nr_dimensions, int *coun
 	if (!res)
 		return throw_oom_error();
 
-	gc_register_finalizer(res, vm_object_finalizer, NULL);
-
-	if (vm_monitor_init(&res->monitor))
-		return throw_internal_error();
+	vm_object_init_common(&res->object);
 
 	res->array_length = counts[0];
-	res->class = class;
+	res->object.class = class;
 
 	if (nr_dimensions == 1)
-		return res;
-
-	struct vm_object **elems = (struct vm_object **) (res + 1);
+		return &res->object;
 
+	struct vm_object **elems = vm_array_elems(&res->object);
 	for (int i = 0; i < counts[0]; ++i)
 		elems[i] = vm_object_alloc_multi_array(elem_class, nr_dimensions - 1, counts + 1);
 
-	return res;
+	return &res->object;
 }
 
 struct vm_object *vm_object_alloc_array(struct vm_class *class, int count)
 {
-	struct vm_object *res;
+	struct vm_array *res;
 
 	if (vm_class_ensure_init(class))
 		return rethrow_exception();
@@ -173,20 +173,29 @@ struct vm_object *vm_object_alloc_array(struct vm_class *class, int count)
 	if (!res)
 		return throw_oom_error();
 
-	gc_register_finalizer(res, vm_object_finalizer, NULL);
-
-	if (vm_monitor_init(&res->monitor))
-		return throw_internal_error();
+	vm_object_init_common(&res->object);
 
 	res->array_length = count;
 
-	struct vm_object **elems = (struct vm_object **) (res + 1);
+	struct vm_object **elems = vm_array_elems(&res->object);
 	for (int i = 0; i < count; ++i)
 		elems[i] = NULL;
 
-	res->class = class;
+	res->object.class = class;
 
-	return res;
+	return &res->object;
+}
+
+struct vm_object *
+vm_object_alloc_array_of(struct vm_class *elem_class, int count)
+{
+	struct vm_class *vmc;
+
+	vmc = vm_class_get_array_class(elem_class);
+	if (!vmc)
+		return throw_internal_error();
+
+	return vm_object_alloc_array(vmc, count);
 }
 
 static struct vm_object *clone_regular(struct vm_object *obj)
@@ -205,7 +214,7 @@ static struct vm_object *clone_array(struct vm_object *obj)
 {
 	struct vm_class *vmc = obj->class;
 	struct vm_class *e_vmc = vmc->array_element_class;
-	int count = obj->array_length;
+	int count = vm_array_length(obj);
 
 	if (e_vmc->kind == VM_CLASS_KIND_PRIMITIVE) {
 		enum vm_type vmtype;
@@ -219,7 +228,7 @@ static struct vm_object *clone_array(struct vm_object *obj)
 		 * object_alloc_primitive_array. */
 		new = vm_object_alloc_primitive_array(type, count);
 		if (new)
-			memcpy(new + 1, obj + 1, vmtype_get_size(vmtype) * count);
+			memcpy(vm_array_elems(new), vm_array_elems(obj), vmtype_get_size(vmtype) * count);
 
 		return new;
 	} else {
@@ -227,7 +236,7 @@ static struct vm_object *clone_array(struct vm_object *obj)
 
 		new = vm_object_alloc_array(vmc, count);
 		if (new)
-			memcpy(new + 1, obj + 1, sizeof(struct vm_object *) * count);
+			memcpy(vm_array_elems(new), vm_array_elems(obj), sizeof(struct vm_object *) * count);
 
 		return new;
 	}
@@ -261,16 +270,6 @@ struct vm_object *vm_object_clone(struct vm_object *obj)
 	return NULL;
 }
 
-void vm_object_lock(struct vm_object *obj)
-{
-	vm_monitor_lock(&obj->monitor);
-}
-
-void vm_object_unlock(struct vm_object *obj)
-{
-	vm_monitor_unlock(&obj->monitor);
-}
-
 struct vm_object *
 vm_object_alloc_string_from_utf8(const uint8_t bytes[], unsigned int length)
 {
@@ -283,7 +282,7 @@ vm_object_alloc_string_from_utf8(const uint8_t bytes[], unsigned int length)
 		return rethrow_exception();
 
 	field_set_int(string, vm_java_lang_String_offset, 0);
-	field_set_int(string, vm_java_lang_String_count, array->array_length);
+	field_set_int(string, vm_java_lang_String_count, vm_array_length(array));
 	field_set_object(string, vm_java_lang_String_value, array);
 
 	return vm_string_intern(string);
@@ -310,40 +309,12 @@ vm_object_alloc_string_from_c(const char *bytes)
 		array_set_field_char(array, i, bytes[i]);
 
 	field_set_int(string, vm_java_lang_String_offset, 0);
-	field_set_int(string, vm_java_lang_String_count, array->array_length);
+	field_set_int(string, vm_java_lang_String_count, vm_array_length(array));
 	field_set_object(string, vm_java_lang_String_value, array);
 
 	return vm_string_intern(string);
 }
 
-typedef void (*exception_init_fn)(struct vm_object *, struct vm_object *);
-
-struct vm_object *new_exception(struct vm_class *vmc, const char *message)
-{
-	struct vm_object *message_str;
-	struct vm_method *mb;
-	struct vm_object *obj;
-
-	obj = vm_object_alloc(vmc);
-	if (!obj)
-		return rethrow_exception();
-
-	if (message == NULL)
-		message_str = NULL;
-	else {
-		message_str = vm_object_alloc_string_from_c(message);
-		if (!message_str)
-			return rethrow_exception();
-	}
-
-	mb = vm_class_get_method(vmc, "<init>", "(Ljava/lang/String;)V");
-	if (!mb)
-		error("constructor not found");
-
-	vm_call_method(mb, obj, message_str);
-	return obj;
-}
-
 bool vm_object_is_instance_of(const struct vm_object *obj,
 	const struct vm_class *class)
 {
@@ -372,9 +343,16 @@ void vm_object_check_array(struct vm_object *obj, jsize index)
 		return;
 	}
 
-	array_len = obj->array_length;
+	array_len = vm_array_length(obj);
 
-	if (index < array_len)
+	/*
+	 * We return if index >= 0 and index < array_len. This can be
+	 * expressed by only one comparison because we are sure that
+	 * array_len is not negative. When index is less than 0 then
+	 * its unsigned representation is greater than any positive
+	 * jint value. Therefore we can skip the explicit check if index >= 0.
+	 */
+	if ((uint32_t)index < (uint32_t)array_len)
 		return;
 
 	sprintf(index_str, "%d > %d", index, array_len - 1);
@@ -410,7 +388,7 @@ void array_store_check(struct vm_object *arrayref, struct vm_object *obj)
 		goto error;
 	}
 
-	err = str_append(str, slash2dots(obj->class->name));
+	err = str_append(str, slash_to_dots(obj->class->name));
 	if (err)
 		goto error;
 
@@ -451,7 +429,7 @@ void vm_object_check_cast(struct vm_object *obj, struct vm_class *class)
 		goto error;
 	}
 
-	err = str_append(str, slash2dots(obj->class->name));
+	err = str_append(str, slash_to_dots(obj->class->name));
 	if (err)
 		goto error;
 
@@ -459,7 +437,7 @@ void vm_object_check_cast(struct vm_object *obj, struct vm_class *class)
 	if (err)
 		goto error;
 
-	err = str_append(str, slash2dots(class->name));
+	err = str_append(str, slash_to_dots(class->name));
 	if (err)
 		goto error;
 
