@@ -24,24 +24,28 @@
  * Please refer to the file LICENSE for details.
  */
 
+#include "arch/memory.h"
+
 #include "vm/call.h"
 #include "vm/class.h"
 #include "vm/die.h"
+#include "vm/errors.h"
 #include "vm/gc.h"
 #include "vm/object.h"
 #include "vm/preload.h"
+#include "vm/reference.h"
 #include "vm/signal.h"
 #include "vm/stdlib.h"
 #include "vm/thread.h"
 
 #include "jit/exception.h"
 
-#include <stdio.h>
+#include <pthread.h>
 #include <stdlib.h>
 #include <errno.h>
-#include <pthread.h>
+#include <stdio.h>
 
-__thread struct vm_exec_env current_exec_env;
+__thread struct vm_exec_env *current_exec_env;
 
 static struct vm_object *main_thread_group;
 
@@ -59,31 +63,54 @@ struct list_head thread_list;
 static bool thread_count_locked;
 static pthread_cond_t thread_count_lock_cond = PTHREAD_COND_INITIALIZER;
 
-static void vm_thread_free(struct vm_thread *thread)
-{
-	vm_free(thread);
-}
-
 static struct vm_thread *vm_thread_alloc(void)
 {
-	struct vm_thread *thread = vm_alloc(sizeof(struct vm_thread));
+	struct vm_thread *thread = malloc(sizeof(struct vm_thread));
 	if (!thread)
 		return NULL;
 
-	if (pthread_mutex_init(&thread->mutex, NULL)) {
-		warn("pthread_mutex_init() failed");
-		vm_thread_free(thread);
-		return NULL;
-	}
+	if (pthread_mutex_init(&thread->mutex, NULL))
+		goto error_mutex;
 
-	thread->state = VM_THREAD_STATE_NEW;
+	atomic_set(&thread->state, VM_THREAD_STATE_NEW);
 	thread->vmthread = NULL;
 	thread->posix_id = -1;
 	thread->interrupted = false;
-	thread->wait_mon = NULL;
+	thread->waiting_mon = NULL;
 	thread->thread_state = THREAD_STATE_CONSISTENT;
+	pthread_cond_init(&thread->park_cond, NULL);
+	pthread_mutex_init(&thread->park_mutex, NULL);
+	thread->unpark_called = false;
+	INIT_LIST_HEAD(&thread->list_node);
 
 	return thread;
+
+ error_mutex:
+	free(thread);
+	return NULL;
+}
+
+static void vm_thread_free(struct vm_thread *thread)
+{
+	pthread_mutex_destroy(&thread->mutex);
+	pthread_mutex_destroy(&thread->park_mutex);
+	pthread_cond_destroy(&thread->park_cond);
+	free(thread);
+}
+
+/*
+ * Called when java.lang.VMThread is being collected.
+ */
+void vm_thread_collect_vmthread(struct vm_object *object)
+{
+	struct vm_thread *thread = vm_thread_from_vmthread(object);
+
+	if (!thread)
+		return;
+
+	assert(thread->ee == NULL);
+
+	vm_thread_free(thread);
 }
 
 /**
@@ -117,9 +144,12 @@ unsigned int vm_nr_threads(void)
 
 void vm_thread_set_state(struct vm_thread *thread, enum vm_thread_state state)
 {
-	pthread_mutex_lock(&thread->mutex);
-	thread->state = state;
-	pthread_mutex_unlock(&thread->mutex);
+	atomic_set(&thread->state, state);
+}
+
+enum vm_thread_state vm_thread_get_state(struct vm_thread *thread)
+{
+	return atomic_read(&thread->state);
 }
 
 static void vm_thread_attach_thread(struct vm_thread *thread)
@@ -176,10 +206,11 @@ int init_threading(void)
 	if (!main_thread)
 		return -ENOMEM;
 
-	main_thread->state = VM_THREAD_STATE_RUNNABLE;
+	atomic_set(&main_thread->state, VM_THREAD_STATE_RUNNABLE);
 	main_thread->vmthread = vmthread;
 	main_thread->posix_id = pthread_self();
 
+	main_thread->ee = vm_get_exec_env();
 	vm_get_exec_env()->thread = main_thread;
 
 	vm_call_method_object(vm_java_lang_Thread_init, thread,
@@ -209,18 +240,57 @@ int init_threading(void)
 	return 0;
 }
 
+static struct vm_exec_env *alloc_exec_env(void)
+{
+	struct vm_exec_env *ee = vm_alloc(sizeof(struct vm_exec_env));
+	if (!ee)
+		return NULL;
+
+	ee->thread = NULL;
+	INIT_LIST_HEAD(&ee->free_monitor_recs);
+	return ee;
+}
+
+static void free_exec_env(struct vm_exec_env *env)
+{
+	struct vm_monitor_record *this, *next;
+
+	struct list_head *list = &env->free_monitor_recs;
+	list_for_each_entry_safe(this, next, list, ee_free_list_node) {
+		vm_monitor_record_free(this);
+	}
+
+	vm_free(env);
+}
+
+void init_exec_env(void)
+{
+	current_exec_env = alloc_exec_env();
+	if (!current_exec_env)
+		error("failed to alloc main execution environment");
+}
+
 /**
  * This is the entry point for all java threads.
  */
 static void *vm_thread_entry(void *arg)
 {
-	struct vm_thread *thread = arg;
+	struct vm_exec_env *ee = arg;
+	struct vm_thread *thread = ee->thread;
 
-	vm_get_exec_env()->thread = thread;
+	current_exec_env = ee;
 
 	setup_signal_handlers();
 	thread_init_exceptions();
 
+	/* XXX: Prevent collection of associated VMThread until
+	 * this method returns. */
+	struct vm_reference *vmthread_ref
+		= vm_reference_alloc_strong(thread->vmthread);
+
+	if (!vmthread_ref)
+		return throw_oom_error();
+
 	vm_call_method(vm_java_lang_VMThread_run, thread->vmthread);
 
 	if (exception_occurred())
@@ -233,6 +303,10 @@ static void *vm_thread_entry(void *arg)
 	vm_thread_detach_thread(vm_thread_self());
 	pthread_mutex_unlock(&threads_mutex);
 
+	thread->ee = NULL;
+	vm_reference_free(vmthread_ref);
+	free_exec_env(ee);
+
 	return NULL;
 }
 
@@ -241,16 +315,41 @@ static void *vm_thread_entry(void *arg)
  */
 int vm_thread_start(struct vm_object *vmthread)
 {
+	/* Force object finalizer execution for vmthread */
+	if (gc_register_finalizer(vmthread, vm_object_finalizer)) {
+		throw_internal_error();
+		return -1;
+	}
+
 	struct vm_thread *thread = vm_thread_alloc();
 	if (!thread) {
-		NOT_IMPLEMENTED;
+		throw_oom_error();
 		return -1;
 	}
 
+	struct vm_exec_env *ee = alloc_exec_env();
+	if (!ee) {
+		throw_oom_error();
+		goto out_free_thread;
+	}
+
+	pthread_attr_t attr;
+
+	if (pthread_attr_init(&attr)) {
+		throw_internal_error();
+		goto out_free_thread;
+	}
+
+	if (pthread_attr_setdetachstate(&attr, true)) {
+		throw_internal_error();
+		goto out_free_thread;
+	}
+
 	/* XXX: no need to lock because @thread is not yet visible to
 	 * other threads. */
 	thread->vmthread = vmthread;
-	thread->state = VM_THREAD_STATE_RUNNABLE;
+
+	atomic_set(&thread->state, VM_THREAD_STATE_RUNNABLE);
 
 	field_set_object(vmthread, vm_java_lang_VMThread_vmdata,
 			 (struct vm_object *) thread);
@@ -267,14 +366,27 @@ int vm_thread_start(struct vm_object *vmthread)
 
 	vm_thread_attach_thread(thread);
 
-	if (pthread_create(&thread->posix_id, NULL, &vm_thread_entry, thread)) {
+	thread->ee = ee;
+	thread->ee->thread = thread;
+
+	if (pthread_create(&thread->posix_id, &attr, &vm_thread_entry, thread->ee)) {
 		vm_thread_detach_thread(thread);
+		thread->ee = NULL;
+		free_exec_env(ee);
 		pthread_mutex_unlock(&threads_mutex);
+
+		signal_new_exception(vm_java_lang_Error, "Unable to create native thread");
 		return -1;
 	}
 
 	pthread_mutex_unlock(&threads_mutex);
+
+	pthread_attr_destroy(&attr);
 	return 0;
+
+ out_free_thread:
+	vm_thread_free(thread);
+	return -1;
 }
 
 void vm_thread_wait_for_non_daemons(void)
@@ -326,26 +438,24 @@ bool vm_thread_interrupted(struct vm_thread *thread)
 
 void vm_thread_interrupt(struct vm_thread *thread)
 {
-	struct vm_monitor *mon;
+	struct vm_object *obj;
+	int err;
 
 	pthread_mutex_lock(&thread->mutex);
 	thread->interrupted = true;
-	mon = thread->wait_mon;
+	obj = thread->waiting_mon;
 	pthread_mutex_unlock(&thread->mutex);
 
-	if (!mon)
+	if (!obj)
 		return;
 
-	/*
-	 * We should wait until thread actually enters a conditional
-	 * wait so that the signal will not be lost. This can
-	 * spuriously wake threads after @thread has left the wait()
-	 * operation but it is fine with the JVM spec.
-	 */
-	while (pthread_mutex_trylock(&mon->mutex))
-		sched_yield();
-	pthread_cond_broadcast(&mon->cond);
-	pthread_mutex_unlock(&mon->mutex);
+	err = vm_object_lock(obj);
+	if (err)
+		return;
+
+	vm_object_notify_all(obj);
+
+	vm_object_unlock(obj);
 }
 
 void vm_lock_thread_count(void)
@@ -367,3 +477,17 @@ void vm_thread_yield(void)
 {
 	sched_yield();
 }
+
+struct vm_thread *vm_thread_from_vmthread(struct vm_object *vmthread)
+{
+	return (struct vm_thread *)field_get_object(vmthread, vm_java_lang_VMThread_vmdata);
+}
+
+struct vm_thread *vm_thread_from_java_thread(struct vm_object *jthread)
+{
+	struct vm_object *vmthread;
+
+	vmthread = field_get_object(jthread, vm_java_lang_Thread_vmThread);
+
+	return vm_thread_from_vmthread(vmthread);
+}
